#include "BitStream.h"
#include "StringCompressor.h"

#include "RakPeerInterface.h"
#include "RakNetworkFactory.h"
#include "RakSleep.h"
#include "MessageIdentifiers.h"
#include "NatPunchthrough.h"

#include <curses.h>
#include <stdlib.h>
#include <string>
#include <iostream>
#include <map>

void print_log(const char* format, ...)
{
	time_t rawtime;
	struct tm * timeinfo;
	time ( &rawtime );
	timeinfo = localtime ( &rawtime );
	printf("%02d-%02d-%d %02d:%02d:%02d\t",timeinfo->tm_mday, 1+timeinfo->tm_mon, 1900+timeinfo->tm_year, timeinfo->tm_hour, timeinfo->tm_min, timeinfo->tm_sec);

	va_list va;
	va_start( va, format );
	vprintf(format, va);
}

NatPunchthrough natPunchthrough;

int main(int argc, char *argv[])
{  
	// Default settings
	SystemAddress facilitatorAddress;
	facilitatorAddress.SetBinaryAddress("83.221.146.11");
	facilitatorAddress.port = 50001;
	int listenPort = 10735;

	// Setup phase
	setlinebuf(stdout);
	std::map<SystemAddress, SystemAddress> addresses;
	RakPeerInterface *peer = RakNetworkFactory::GetRakPeerInterface();
		
	if (argc > 1)
		listenPort = atoi(argv[1]);
	print_log("Connection Tester port set to %d\n",listenPort);
		
	natPunchthrough.FacilitateConnections(true);

	peer->SetMaximumIncomingConnections(1000);
	SocketDescriptor sd(listenPort,0);
	peer->Startup(1000, 0, &sd, 1);
	peer->AttachPlugin(&natPunchthrough);
	
	// Connect to facilitator
	if (!peer->Connect(facilitatorAddress.ToString(false), facilitatorAddress.port,0,0))
	{
		print_log("Error: Failed to connect to NAT facilitator at %s\n", facilitatorAddress.ToString());
		return 1;
	}
	
	Packet *packet;
	while (1)
	{
		packet=peer->Receive();
		while (packet)
		{
			switch (packet->data[0])
			{
				case ID_DISCONNECTION_NOTIFICATION:
					print_log("%s has diconnected\n", packet->systemAddress.ToString());
					break;
				case ID_CONNECTION_LOST:
					print_log("Connection to %s lost\n", packet->systemAddress.ToString());
					if (packet->systemAddress == facilitatorAddress)
					{
						print_log("Reconnecting in 10 seconds\n");
						sleep(10);
						if (!peer->Connect(facilitatorAddress.ToString(false), facilitatorAddress.port,0,0))
						{
							print_log("Error: Failed to connect to NAT facilitator at %s\n", facilitatorAddress.ToString());
							return 1;
						}
					}
					break;
				case ID_NEW_INCOMING_CONNECTION:
					print_log("New connection established to %s\n", packet->systemAddress.ToString());
					break;
				case ID_CONNECTION_REQUEST_ACCEPTED:
					print_log("Connected to %s\n", packet->systemAddress.ToString());
					// Immediately disconnect again if this is a connection test
					if (packet->systemAddress != facilitatorAddress)
						peer->CloseConnection(packet->systemAddress, true, 0);
					break;
				case ID_CONNECTION_ATTEMPT_FAILED:
				{
					if (addresses.count(packet->systemAddress) > 0)
					{
						RakNet::BitStream bitStream;
						bitStream.Write((unsigned char)254);
						peer->Send(&bitStream, HIGH_PRIORITY, RELIABLE, 0, addresses[packet->systemAddress], false);
						addresses.erase(packet->systemAddress);
					}
					else
					{
						print_log("Failed to connect to %s", packet->systemAddress.ToString());
					}
					break;
				}
				case ID_NAT_TARGET_NOT_CONNECTED:
				{
					SystemAddress systemAddress;
					RakNet::BitStream bitStream(packet->data, packet->length, false);
					bitStream.IgnoreBits(8); // Ignore the ID_...
					bitStream.Read(systemAddress);
					print_log("NAT target %s is not connected to the facilitator\n", systemAddress.ToString());
					break;
				}
				case ID_NAT_TARGET_CONNECTION_LOST:
				{
					SystemAddress systemAddress;
					RakNet::BitStream bitStream(packet->data, packet->length, false);
					bitStream.IgnoreBits(8); // Ignore the ID_...
					bitStream.Read(systemAddress);
					print_log("NAT connection to %s lost\n", systemAddress.ToString());
					break;
				}
				case 253:
				{
					int remotePort;
					RakNet::BitStream bitStream(packet->data, packet->length, false);
					bitStream.IgnoreBits(8); // Ignore the ID_...
					bitStream.Read(remotePort);

					// Add system to address list so it can be told if the test failed
					SystemAddress remoteTarget;
					remoteTarget.binaryAddress = packet->systemAddress.binaryAddress;
					remoteTarget.port = remotePort;
					addresses[remoteTarget] = packet->systemAddress;
					char remote[32];
					strcpy(remote, remoteTarget.ToString());
					print_log("Adding %s to %s in map\n", packet->systemAddress.ToString(), remote);

					print_log("Attempting connection test to %s:%d\n", packet->systemAddress.ToString(false), remotePort);
					if (!peer->Connect(packet->systemAddress.ToString(false), remotePort, 0, 0))
						print_log("Failed to connect to %s:%d\n", packet->systemAddress.ToString(false), remotePort);					
					break;
				}
				default:
					print_log("Unknown ID %d from %s\n", packet->data[0], packet->systemAddress.ToString());
			}
			peer->DeallocatePacket(packet);
			packet=peer->Receive();
		}
		RakSleep(30);
	}

	peer->Shutdown(100,0);
	RakNetworkFactory::DestroyRakPeerInterface(peer);
	print_log("Quitting\n");
			
	return 0;
}
